/*
 * snowflake.c
 *
 *  Created on: 13-01-2012
 *      Author: Andrzej Michalec (amichalec.net)
 */

#include <stddef.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include "bitmacros.h"
#include "snowflake.h"

#define PC_R			PC0
#define PC_G			PC1
#define PC_B			PC2
#define PC_SEG_A		PC3
#define PC_SEG_B		PC4
#define PC_SEG_C		PC5
#define PD_SHIFT_CLEAR	PD4
#define PD_SHIFT_CLOCK	PD5
#define PD_BUFF_WRITE	PD6
#define PD_BUFF_DISABLE	PD7

#define EFFECT_TIMER_RESOLUTION 	10	//in milliseconds
/**
 * Structure for current LED light state.
 * If effect is set either eff_data or eff_iter is not-null.
 */
typedef struct pwm_light {
	uint16_t brightness; //current brightness (static & effect)
	struct light_effect * eff_data; //effect played from data, or null for static light
	uint16_t eff_step; //index in eff_data or step in eff_iter
	uint16_t eff_time_left; //count-down to change to next effect
	effect_iterator_t eff_iter; //effect played from iterator
} pwm_light_t;

/**
 * PWM phases indexed by brightness level.
 */
const int g_pwm_phases[SF_MAX_BRIGHTNESS + 1] = { 0, 1, 3, 6, 10, 18, 32, 48,
		64, 76, 106, 132, 164, 192, 224, 255 };
// Current phase
volatile uint8_t g_pwm_phase;

// LED on/off bits for PWM cycle
// 8 bytes for 15 brightness levels, each byte represents
// PORTC state (RGB plus segments)
volatile uint8_t g_pwm_cache[15][8];

effect_finished_handler_t g_effect_listener;

// Lights 0..23 are white, 24th is RGB
volatile pwm_light_t g_ligths[SF_MAX_LIGHTS + 1];

void sf_init() {
	PORTC = 0x00;
	DDRC = 0xFF;
	PORTD = 0x00;
	DDRD = 0xF0; //4 older bits only

	// cleaning 74hc595 registers
	CLEAR_BIT(PORTD, PD_SHIFT_CLEAR);
	SET_BIT(PORTD, PD_SHIFT_CLEAR);
	SET_BIT(PORTD, PD_BUFF_WRITE);
	CLEAR_BIT(PORTD, PD_BUFF_WRITE);

	for (uint8_t i = 0; i < SF_MAX_LIGHTS; i++) {
		g_ligths[i].brightness = 0;
	}

	SET_MASK(TIMSK, BIT(TOIE0) | BIT(TOIE2)); // enable 2 timer interruptions
	//SET_MASK(TCCR0, BIT(CS00) | BIT(CS01)); // f/64
	SET_MASK(TCCR0, BIT(CS02)); // f/256
	SET_MASK(TCCR2, BIT(CS22) | BIT(CS21) | BIT(CS20)); // f/1024
	TCNT0 = 0xFF;
	TCNT2 = 0xFF;
	sei();
}

/**
 * Push bytes to serial registers and to RGB diode.
 */
void __push_pwm_cache() {
	const uint8_t segmask = 0b00111000;
	const uint8_t rgbmask = 0b00000111;
	const uint8_t invsegmask = ~segmask;
	const uint8_t invrgbmask = ~rgbmask;
	if (g_pwm_phase == 0) {
		// zero segment bits
		PORTC &= invsegmask;
		for (uint8_t i = 0; i < 8; i++) {
			// clock cycle for shift register
			SET_BIT(PORTD, PD_SHIFT_CLOCK);
			CLEAR_BIT(PORTD, PD_SHIFT_CLOCK);
		}
	} else {
		for (uint8_t i = 0; i < 8; i++) {
			// zero segment bits only
			PORTC &= invsegmask;
			// set segment bits
			PORTC |= (segmask & g_pwm_cache[g_pwm_phase - 1][i]);
			// clock cycle for shift register
			SET_BIT(PORTD, PD_SHIFT_CLOCK);
			CLEAR_BIT(PORTD, PD_SHIFT_CLOCK);
		}
	}
	// clock cycle for writing from shift to buffer register
	SET_BIT(PORTD, PD_BUFF_WRITE);
	CLEAR_BIT(PORTD, PD_BUFF_WRITE);
	// zero RGB bits and then set their new values
	// (skip g_pwm_phase=15 since g_pwm_cache is indexed 0..14)
	if (g_pwm_phase < SF_MAX_BRIGHTNESS) {
		uint8_t newc = PORTC;
		newc &= invrgbmask;
		newc |= (rgbmask & g_pwm_cache[g_pwm_phase][0]);
		PORTC = newc;
	}
}

/**
 * For light in 0..23 brightness is 0..15 (for white lights)
 * For light equals to SF_RGB_LIGHT_NUMBER brightness is packed as in SF_RGB() macro.
 */
void __update_pwm_cache(uint8_t light, uint16_t brightness) {
	if (light != SF_RGB_LIGHT_NUMBER) {
		// updating white LEDs
		const int segpin[3] = { PC_SEG_A, PC_SEG_B, PC_SEG_C };
		uint8_t seg = light / 8;
		uint8_t byte = 7 - (light % 8);
		for (uint8_t i = 0; i < SF_MAX_BRIGHTNESS; i++) {
			if (i < brightness) {
				SET_BIT(g_pwm_cache[i][byte], segpin[seg]);
			} else {
				CLEAR_BIT(g_pwm_cache[i][byte], segpin[seg]);
			}
		}
	} else {
		// updating RGB LED
		const uint8_t rgbpin[] = { PC_R, PC_G, PC_B };
		uint8_t rgb[3];
		rgb[0] = SF_R(brightness);
		rgb[1] = SF_G(brightness);
		rgb[2] = SF_B(brightness);
		for (uint16_t color = 0; color < 3; color++) {
			for (uint16_t i = 0; i < SF_MAX_BRIGHTNESS; i++) {
				if (i < rgb[color]) {
					SET_BIT(g_pwm_cache[i][0], rgbpin[color]);
				} else {
					CLEAR_BIT(g_pwm_cache[i][0], rgbpin[color]);
				}
			}
		}
	}
}

/**
 * LED PWM modulation.
 */ISR( TIMER0_OVF_vect)
{
	TCNT0 = 0xFF - g_pwm_phases[g_pwm_phase];
	__push_pwm_cache();
	g_pwm_phase = (g_pwm_phase + 1) % (SF_MAX_BRIGHTNESS + 1);
}

volatile uint8_t g_iterRecur = 0;
void __effect_clock_tick(volatile struct pwm_light * pl, uint8_t light_number) {
	if (pl->eff_data != NULL) {
		if (pl->eff_time_left < EFFECT_TIMER_RESOLUTION) {
			if (pl->eff_step >= pl->eff_data->length) {
				pl->eff_step = 0;
				if (pl->eff_data->next_effect != NULL) {
					// relinking to next effect
					struct light_effect * next = pl->eff_data->next_effect;
					pl->eff_data = next;
					pl->eff_time_left = *(pl->eff_data->time_tab);
					pl->brightness = *(pl->eff_data->brightness_tab);
					__update_pwm_cache(light_number, pl->brightness);
				} else {
					// end of effects, cleaning & calling listener
					pl->eff_data = NULL;
					pl->brightness = 0;
					__update_pwm_cache(light_number, pl->brightness);
					if (g_effect_listener != NULL) {
						g_effect_listener(light_number);
					}
				}
			} else {
				pl->brightness = *(pl->eff_data->brightness_tab + pl->eff_step);
				__update_pwm_cache(light_number, pl->brightness);
				pl->eff_time_left = *(pl->eff_data->time_tab + pl->eff_step);
			}
			pl->eff_step++;
		} else {
			pl->eff_time_left -= EFFECT_TIMER_RESOLUTION;
		}
	} else if (pl->eff_iter != NULL) {
		if (pl->eff_time_left < EFFECT_TIMER_RESOLUTION) {
			// callback to get new brightness and time
			// with recursion prevention (does it require ATOMIC_BLOCK?)
			if (!g_iterRecur) {
				g_iterRecur = 1;
				(*pl->eff_iter)(light_number, (uint8_t) pl->eff_step,
						(uint16_t*) &pl->brightness,
						(uint16_t*) &pl->eff_time_left);
				g_iterRecur = 0;
			}
			__update_pwm_cache(light_number, pl->brightness);
			pl->eff_step++;
		} else {
			pl->eff_time_left -= EFFECT_TIMER_RESOLUTION;
		}
	}
}

/**
 * LED effect timer
 */ISR( TIMER2_OVF_vect)
{
	TCNT2 = 0xFF - 15 * EFFECT_TIMER_RESOLUTION;
	volatile struct pwm_light * pl;
	//max light plus one for RGB light!
	for (uint8_t i = 0; i < SF_MAX_LIGHTS + 1; i++) {
		pl = &(g_ligths[i]);
		__effect_clock_tick(pl, i);
	}
}

uint8_t __check_effect(light_effect_t * effect) {
	if (effect != NULL) {
		if (effect->brightness_tab == NULL || effect->time_tab == NULL)
			return 1;
		if (effect->length < 1)
			return 1;
	}
	return 0;
}

void __set_light(uint8_t light_type, uint8_t light_number, uint16_t brightness,
		light_effect_t * eff_data, effect_iterator_t eff_iter) {
	if (light_type == SF_LIGHT_TYPE_POINT) {
		if (light_number > SF_MAX_LIGHTS) {
			return;
		}
		if (__check_effect(eff_data)) {
			return;
		}
		volatile pwm_light_t * pl = &g_ligths[light_number];
		if (eff_data != NULL) {
			//disable effect IRQ processing while messing up with effect structure
			CLEAR_BIT(TIMSK, TOIE2);
			pl->brightness = 0;
			pl->eff_time_left = 0;
			pl->eff_step = 0;
			pl->eff_data = eff_data;
			pl->eff_iter = NULL;
			__effect_clock_tick(pl, light_number);
			__update_pwm_cache(light_number, brightness);
			SET_BIT(TIMSK, TOIE2);
		} else if (eff_iter != NULL) {
			CLEAR_BIT(TIMSK, TOIE2);

			pl->brightness = 0;
			pl->eff_time_left = 0;
			pl->eff_step = 0;
			pl->eff_data = NULL;
			pl->eff_iter = eff_iter;
			__effect_clock_tick(pl, light_number);
			__update_pwm_cache(light_number, brightness);
			SET_BIT(TIMSK, TOIE2);
		} else {
			if (light_number != SF_RGB_LIGHT_NUMBER && brightness
					> SF_MAX_BRIGHTNESS) {
				return;
			}
			if (light_number == SF_RGB_LIGHT_NUMBER && brightness > 0x0FFF) {
				return;
			}
			CLEAR_BIT(TIMSK, TOIE2);
			pl->eff_data = NULL;
			pl->eff_iter = NULL;
			pl->brightness = brightness;
			pl->eff_step = 0;
			pl->eff_time_left = 0;
			__update_pwm_cache(light_number, brightness);
			SET_BIT(TIMSK, TOIE2);
		}
	} else if (light_type == SF_LIGHT_TYPE_ARM) {
		for (uint8_t i = 0; i < SF_MAX_CIRCLES; i++) {
			__set_light(SF_LIGHT_TYPE_POINT, i + SF_MAX_CIRCLES * light_number,
					brightness, eff_data, eff_iter);
		}
	} else if (light_type == SF_LIGHT_TYPE_CIRCLE) {
		uint8_t light = SF_MAX_CIRCLES - 1 - light_number;
		for (uint8_t i = 0; i < SF_MAX_ARMS; i++) {
			__set_light(SF_LIGHT_TYPE_POINT, light, brightness, eff_data,
					eff_iter);
			light += SF_MAX_CIRCLES;
		}
	}
}

void sf_set_light(uint8_t light_type, uint8_t light_number, uint16_t brightness) {
	__set_light(light_type, light_number, brightness, NULL, NULL);
}

uint16_t sf_get_light(uint8_t light_number) {
	if (light_number > SF_MAX_LIGHTS) {
		return 0;
	}
	return g_ligths[light_number].brightness;
}

void sf_set_light_effect_data(uint8_t light_type, uint8_t light_number,
		light_effect_t * effect) {
	__set_light(light_type, light_number, 0, effect, NULL);
}

void sf_set_light_effect_iterator(uint8_t light_type, uint8_t light_number,
		effect_iterator_t iterator) {
	__set_light(light_type, light_number, 0, NULL, iterator);
}

void sf_set_listener(effect_finished_handler_t func) {
	g_effect_listener = func;
}

uint8_t sf_is_effect_in_progress(uint8_t light) {
	if (light > SF_MAX_LIGHTS) {
		return 0;
	}
	return (g_ligths[light].eff_data || g_ligths[light].eff_iter ? 1 : 0);
}

light_effect_t * sf_allocate(uint8_t length) {
	light_effect_t * pe;
	// protecting thread unsafe malloc
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		pe = malloc(sizeof *pe);
	}
	pe->length = length;
	if (length > 0) {
		ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
			pe->brightness_tab = malloc(sizeof pe->brightness_tab * length);
			pe->time_tab = malloc(sizeof pe->time_tab * length);
		}
	} else {
		pe->brightness_tab = NULL;
		pe->time_tab = NULL;
	}
	pe->next_effect = NULL;
	return pe;
}

void sf_free(light_effect_t * effect) {
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		free(effect->brightness_tab);
		free(effect->time_tab);
		free(effect);
	}
}

light_effect_t * sf_white2rgb(light_effect_t * eff) {
	if (__check_effect(eff)) {
		return NULL;
	}
	light_effect_t * ret = sf_allocate(eff->length);
	ret->next_effect = NULL;
	for (uint16_t i = 0; i < eff->length; i++) {
		*(ret->time_tab + i) = *(eff->time_tab + i);
		uint16_t br = *(eff->brightness_tab + i);
		//same brightness for every component
		*(ret->brightness_tab + i) = SF_RGB(br, br, br);
	}
	return ret;
}

light_effect_t * sf_rgb2white(light_effect_t * eff) {
	if (__check_effect(eff)) {
		return NULL;
	}
	light_effect_t * ret = sf_allocate(eff->length);
	ret->next_effect = NULL;
	for (uint8_t i = 0; i < eff->length; i++) {
		*(ret->time_tab + i) = *(eff->time_tab + i);
		uint16_t br = *(eff->brightness_tab + i);
		//average brightness of components
		uint16_t avg = (SF_R(br) + SF_G(br) + SF_B(br)) / 3;
		*(ret->brightness_tab + i) = avg;
	}
	return ret;
}
