/*
 * snowflake.h
 *
 *  Created on: 13-01-2012
 *      Author: Andrzej Michalec (amichalec.net)
 *
 *
 * Snowflake is set of 24 white lights on 6-arms snowflake
 * with RGB light in the center as depicted below:
 *
 * 20          0           4
 *    21       1        5
 *       22    2     6
 *          23 3  7
 *            rgb
 *          19 15 11
 *       18    14    10
 *    17       13       9
 * 16          12          8
 *
 * light - one of 24 single white LED lights located
 * 		on front side of PCB. Lights are numbered 0..23
 *
 * arm - snowflake arm of lights. Arms are numbered 0..5
 * 		countered clockwise. Each arm has 4 lights
 * 		e.g. arm 4 has lights 16, 17, 18 & 19.
 *
 * circle - circle of lights with same radius from center.
 * 		Circles are numbered from 0..3 counted from the
 * 		center outside. Each circle has 6 lights e.g.
 * 		circle 0 has lights 3, 7, 11, 15, 19 & 23
 *
 * brightness - level of LED light intensity.
 *
 * light effect - description of how to change brightness
 * 		of given light over time to achieve pulsing, glimmering etc.
 * 		Effect can be described in two ways: precomputed data or
 * 		iterator function
 *
 * 			* data - client must provide data structure
 * 				with array containing full effects. Each position in array
 * 				describes value of brightness and time (in milliseconds)
 * 				how long to keep that brightness.
 *
 * 			* iterator - client must provided callback function that will
 * 				compute next brightness and time.
 */

#ifndef SNOWFLAKE_H_
#define SNOWFLAKE_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#define SF_MAX_LIGHTS			24
#define SF_MAX_ARMS				6
#define SF_MAX_CIRCLES			4
#define SF_MAX_BRIGHTNESS		15

#define SF_LIGHT_TYPE_POINT		0
#define SF_LIGHT_TYPE_ARM		1
#define SF_LIGHT_TYPE_CIRCLE	2

#define SF_RGB_LIGHT_NUMBER		24

#define SF_RGB(r, g, b) ( (uint16_t)( ((0x000F & r) << 8) | ((0x000F & g) << 4) | (0x000F & b)) )
#define SF_R(word) ((0x0F00 & word) >> 8)
#define SF_G(word) ((0x00F0 & word) >> 4)
#define SF_B(word) (0x000F & word)

/**
 * Light effect structure.
 *
 * length - size of brightness and time arrays (number of steps).
 * brightnessTab - array of brightness 0..15 for each step
 * timeTab - array of time in milliseconds to spend in each step
 * nextEffect - automatic progression with next effect or NULL
 * 		to stop after effect
 */
typedef struct light_effect {
	uint8_t length;
	uint16_t *brightness_tab;
	uint16_t *time_tab;
	struct light_effect * next_effect;
} light_effect_t;

/**
 * Definition of finished effect callback (for data-oriented effects).
 * When effect for given light is played to the end, client callback
 * handler is called passing this light number as parameter to allow client
 * set new effect in continuous manner.
 */
typedef void (*effect_finished_handler_t)(uint8_t ligth_of_finished_effect);

/**
 * Definition of effect iterator callback method (for calculation-oriented
 * effects). When new brightness and time must be computed for given for next
 * step, client callback handler is called passing light number and step number.
 * Client is responsible to write new_brightness and new_time. When called
 * iterator can unregister iterator to stop effect calling
 * "sf_set_light_effect_iterator(light, step, NULL)" or register other callback;
 * new iterator state will be in effect in next step so that setting new_time
 * is important.
 */
typedef void (*effect_iterator_t)(uint8_t light, uint8_t step,
		uint16_t * new_brightness, uint16_t * new_time);

/**
 * System initialization. Registers ISRs to refresh light state and
 * replay effects.
 */
extern void sf_init();

/**
 * Light with constant brightness.
 *
 * 	@param light_type - type of light, SF_LIGHT_POINT for single light,
 * 		SF_LIGHT_ARM for arm of lights and SF_LIGHT_CIRCLe for circle
 * 		of lights.
 * 	@param light_number - zero-indexed number of light point, arm or circle
 * 		as specified by light_type.	When central light SF_RGB_LIGHT_NUMBER is
 * 		specified, light_type must be SF_LIGHT_POINT.
 * 	@param brightness - For white light value from 0 to 15, zero to turn off,
 * 		15 for maximum light. For RGB light composites have to be encoded by
 * 		SF_RGB() macrodefinition.
 *
 * This method cancels currently replayed effect (if any). Incorrect value
 * of parameter causes operation to be skipped w/o side effect.
 */
extern void sf_set_light(uint8_t light_type, uint8_t light_number,
		uint16_t brightness);

/**
 * Reads current light brightness. Light number is alway point LED. For
 * SF_RGB_LIGHT_NUMBER color triplet is returned to be decoded by SF_R(),
 * SF_G() and SF_B() macrodefinitions.
 */
extern uint16_t sf_get_light(uint8_t light_number);

/**
 * Set effect to replay on particular light using precomputed data structure.
 *
 * 	@param light_type - type of light, SF_LIGHT_POINT for single light,
 * 		SF_LIGHT_ARM for arm of lights and SF_LIGHT_CIRCLe for circle
 * 		of lights.
 * 	@param light_number - zero-indexed number of light point, arm or circle
 * 		as specified by light_type.	When central light SF_RGB_LIGHT_NUMBER is
 * 		specified, light_type must be SF_LIGHT_POINT.
 * 	@param effect - new effect to replay; see light_effect_t structure
 * 		for details.
 *
 * This method cancels currently replayed effect (if any). Incorrect value
 * of parameter causes operation to be skipped w/o side effect.
 */
extern void sf_set_light_effect_data(uint8_t light_type, uint8_t light_number,
		light_effect_t * effect);

/**
 * Set effect to replay on particular light using callback function.
 *
 * 	@param light_type - type of light, SF_LIGHT_POINT for single light,
 * 		SF_LIGHT_ARM for arm of lights and SF_LIGHT_CIRCLe for circle
 * 		of lights.
 * 	@param light_number - zero-indexed number of light point, arm or circle
 * 		as specified by light_type.	When central light SF_RGB_LIGHT_NUMBER
 * 		specified, light_type is ignored.
 * 	@param iterator - function called to compute new step of effect; see
 * 		effect_iterator_t definition for details.
 *
 * This method cancels currently replayed effect (if any). Incorrect value
 * of parameter causes operation to be skipped w/o side effect.
 */
extern void sf_set_light_effect_iterator(uint8_t light_type,
		uint8_t light_number, effect_iterator_t iterator);

/**
 * Registers effect finalization callback function.
 *
 * @param func - function called each time when light effect ends; see
 * 		effect_finished_handler_t definition for details.
 */
extern void sf_set_listener(effect_finished_handler_t func);

/**
 * Checks light effect for lights 0..SF_MAX_LIGHTS or SF_RGB_LIGHT_NUMBER.
 * Returns zero if no effect is associated, non-zero otherwise.
 */
extern uint8_t sf_is_effect_in_progress(uint8_t light_number);

/**
 * Allocates and free light effect structure (convenience method).
 */
extern light_effect_t * sf_allocate(uint8_t length);
extern void sf_free(light_effect_t * effect);

/**
 * Converts brightness data of effect structure from white into
 * RGB and back (white diode codes only value, while RGB codes 4 bits of
 * color in 2 bytes). Passed data structure is not changed, returned
 * data structure is deep copy of passed data. If passed data structure
 * is invalid NULL is returned.
 */
extern light_effect_t * sf_white2rgb(light_effect_t * whiteEffect);
extern light_effect_t * sf_rgb2white(light_effect_t * rgbEffect);

#ifdef __cplusplus
}
#endif

#endif /* SNOWFLAKE_H_ */
